class RestrictionError(Exception):
    """
    Exceptions for Restrictions.
    """

    @classmethod
    def bad_data(cls, data, allowed):
        """
        Run time error. Data violated restrictions.
        """
        return cls("'%s' not allowed per restriction '%s'" % (data, allowed))

    @classmethod
    def from_mixed_value_and_type(cls, allowed):
        """
        Compile time error. Restriction syntax not allowed.
        """
        return cls("Mixed type and value restrictions unsupported '%s'" % allowed)

    @classmethod
    def from_unsupported_dataobj_in_rstr_list(cls, allowed):
        """
        Compile time error. DO in restriction can only be combined with None type.
        """
        return cls("DataObject in restriction list '%s' only valid with type(None)." % allowed)

    @classmethod
    def from_dataobj_in_rstr_list(cls, allowed):
        """
        Compile time error. DO in restriction cannot be combined other data types (expect None type).
        """
        return cls("DataObjects not allowed in restriction lists '%s'. Use (<DataObject>, <default_value>)" % allowed)

    @classmethod
    def from_invalid_default_value(cls, default):
        """
        Compile time error. Default value for restriction is invalid.
        """
        return cls('Invalid default "%s".' % default)

    @classmethod
    def from_unsupported(cls, allowed):
        """
        Compile time error. Restriction syntax is incorrect.
        """
        return cls("Malformed restriction. Allowed '%s' is of type '%s'." % (allowed, type(allowed)))



"""
Houses the commonly used managed restriction, MgdDatetime.
:date_created: 2020-06-28
"""

from datetime import date, datetime

from do_py.common import R
from do_py.data_object.restriction import ManagedRestrictions
from do_py.exceptions import RestrictionError


class MgdDatetime(ManagedRestrictions):
    """
    Managed from and to date/datetime restrictions. It follows the following logic:

                                        From                        To
    None                                Epoch date(time)            Epoch date(time)
    Valid Datetime/Date instance        Do nothing                  Do nothing
    Everything else                     Parse as ISO date(time)     Parse as ISO date(time)

    Note:
        The default value will not be set properly in `strict=False` initializations that are missing the relevant key.
        This is due to strictness impacting ManagedRestrictions execution; the `manage` method does not execute on
        `strict=False`. Since this implementation sets the default in this method, the abstraction fails to
    Example:
        class A(DataObject):
            _restrictions = {
                'from_date': MgdDatetime.from_from_date()
                }

        A(strict=False)  # {"from_date": null}
        A(data={'from_date': None}, strict=False)  # {"from_date": "1969-12-31"}
        A({'from_date': None})  # {"from_date": "1969-12-31"}
    """
    dt_obj = None
    _restriction = R()
    _parse_dt_fmt = {
        datetime: '%Y-%m-%dT%H:%M:%S',
        date: '%Y-%m-%d'
        }
    defaults = {
        'from': lambda dt: dt.fromtimestamp(0),
        'to': lambda dt: dt.now() if dt is datetime else dt.today()
        }

    def __init__(self, dt_obj=None, default_key=None, nullable=False, *args, **kwargs):
        """
        :param dt_obj: Initialize the restriction as date or datetime.
        :type dt_obj: Type[Union[datetime, date]]
        :param default_key: Manages "from" or "to"
        :type default_key: str
        """
        assert dt_obj in self._parse_dt_fmt, 'Invalid "dt_obj"(=%s)' % dt_obj
        assert default_key is None or default_key in self.defaults, 'Invalid "default_key"(=%s)' % default_key
        self.dt_obj = dt_obj
        self.default_key = default_key
        self.nullable = nullable
        self._restriction = R.DATETIME if self.dt_obj is datetime else R.DATE
        super(MgdDatetime, self).__init__(*args, **kwargs)

    def manage(self):
        """
        Implements the logic outlined in class docstring.
        Uses datetime.strptime by design to be more strict on the string parsing for ISO format.
        """
        if self.data is None:
            if self.default_key:
                self.data = self.defaults[self.default_key](self.dt_obj)
            elif not self.nullable:
                raise RestrictionError.bad_data(self.data, self._restriction.allowed)
        elif type(self.data) not in [datetime, date]:
            self.data = datetime.strptime(self.data, self._parse_dt_fmt[self.dt_obj])
            if self.dt_obj is date:
                self.data = self.data.date()

        if self.data is not None and self.dt_obj is datetime:
            self.data = self.data.replace(microsecond=0)

    @classmethod
    def from_from_date(cls):
        """
        Create a MgdDatetime instance for validating a `from_date` restriction.
        This will validate a DATE format, not DATETIME.
        :rtype: MgdDatetime
        """
        return cls(dt_obj=date, default_key='from')

    @classmethod
    def from_to_date(cls):
        """
        Create a MgdDatetime instance for validating a `to_date` restriction.
        This will validate a DATE format, not DATETIME.
        :rtype: MgdDatetime
        """
        return cls(dt_obj=date, default_key='to')

    @classmethod
    def from_from_datetime(cls):
        """
        Create a MgdDatetime instance for validating a `from_datetime` restriction.
        This will validate a DATETIME format, not DATE.
        :rtype: MgdDatetime
        """
        return cls(dt_obj=datetime, default_key='from')

    @classmethod
    def from_to_datetime(cls):
        """
        Create a MgdDatetime instance for validating a `to_datetime`
        This will validate a DATETIME format, not DATE.
        :rtype: MgdDatetime
        """
        return cls(dt_obj=datetime, default_key='to')

    @classmethod
    def datetime(cls):
        """
        Create MgdDatetime instance for validating and standardizing a DATETIME format.
        :rtype: MgdDatetime
        """
        return cls(dt_obj=datetime)

    @classmethod
    def date(cls):
        """
        Create MgdDatetime instance for validating and standardizing a DATE format.
        :rtype: MgdDate
        """
        return cls(dt_obj=date)

    @classmethod
    def null_date(cls):
        """
        Create MgdDatetime instance for validating and standardizing a DATE format that is nullable.
        :rtype: MgdDate
        """
        return cls(dt_obj=date, nullable=True)



"""
Useful methods, classes, and tools.
An important note on utils is that they should have no dependencies on the rest of the project.
"""
from .json_encoder import MyJSONEncoder
from .properties import cached_property, classproperty, is_cached_property, is_classmethod, is_property



"""
Commonly used restrictions.
:date_created: 2020-06-28
"""
import sys
from datetime import date, datetime

from future.types import newint, newlist, newstr
from past.builtins import long, unicode

from do_py.data_object import Restriction
from do_py.utils import classproperty


class R(object):
    """
    The definition of a restriction for DataObjects:
    1. Allowed types or values.
       This is required for all restrictions and sets a strong standard for value and type checking.
       Value checks are not limited to a particular data type.
    2. Optionally, a default value for a restriction may be defined. This allows us to initiate DataObjects without
       providing data for this key/restriction.
       The primary use case for default values is to fill in gaps in data, i.e. front-end requests, 3rd party APIs, etc.

    Syntax:
        R(<restriction 1>, <restriction 2>, ..., default=<default value>)
        *args for R define either the types or values allowed (types and values may NOT be mixed).
        The keyword `default` defines the optional default value for the restriction.

    Usages:
        R(int, float)  # Allows integers and floats data. No default provided.
        R('hello', 'world')  # Allows the strings 'hello' and 'world' to be used as data values. No default provided.
        R(int, float, default=1)  # Again, allows integers and floats, but 1 is default value when no data is provided.

    Example:
        class A(DataObject):
            _restrictions = {
                'x': R(int),
                'y': R(int, default=12)
                }

        a = A(strict=False)
        a  # {'x': null, 'y': 12}
    """

    def __new__(cls, *args, **kwargs):
        """
        :rtype: Restriction
        """
        return Restriction(list(args), **kwargs)

    def __init__(self, *args, **kwargs):
        """
        :param args: Values or types for restriction definition
        :param kwargs: optional arguments for restriction instantiation
        :keyword default: default restriction value
        """

    def __call__(self, *args, **kwargs):
        """
        Dummy to support classproperty
        """

    @classproperty
    def INT(cls):
        """
        Shortcut for an int restriction.
        :rtype: Restriction
        """
        if sys.version_info.major == 3:
            return cls(int, newint.newint)
        return cls(int, newint)

    @classproperty
    def FLOAT(cls):
        """
        Shortcut for a float restriction.
        :rtype: Restriction
        """
        return cls(float)

    @classproperty
    def STR(cls):
        """
        Shortcut for string restriction.
        :rtype: Restriction
        """
        if sys.version_info.major == 3:
            return cls(str, unicode, newstr.newstr)
        return cls(str, unicode, newstr)

    @classproperty
    def NULL_INT(cls):
        """
        Shortcut for a nullable int restriction.
        :rtype: Restriction
        """
        if sys.version_info.major == 3:
            return cls(int, newint.newint, type(None))
        return cls(int, newint, type(None))

    @classproperty
    def NULL_FLOAT(cls):
        """
        Shortcut for a nullable float restriction.
        :rtype: Restriction
        """
        return cls(float, type(None))

    @classproperty
    def NULL_STR(cls):
        """
        Shortcut for a nullable strinng restriction.
        :rtype: Restriction
        """
        if sys.version_info.major == 3:
            return cls(str, unicode, newstr.newstr, type(None))
        return cls(str, unicode, newstr, type(None))

    @classproperty
    def LIST(cls):
        """
        Shortcut for a list restriction.
        :rtype: Restriction
        """
        if sys.version_info.major == 3:
            return cls(list, newlist.newlist)
        return cls(list, newlist)

    @classproperty
    def SET(cls):
        """
        Shortcut for a set restriction.
        :rtype: Restriction
        """
        return cls(set)

    @classproperty
    def NULL_LIST(cls):
        """
        Shortcut for a nullable list restriction.
        :rtype: Restriction
        """
        if sys.version_info.major == 3:
            return cls(list, newlist.newlist, type(None))
        return cls(list, newlist, type(None))

    @classproperty
    def BOOL(cls):
        """
        Shortcut for a bool restriction.
        :rtype: Restriction
        """
        return cls(bool)

    @classproperty
    def DATETIME(cls):
        """
        Shortcut for a datetime.datetime restriction.
        :rtype: Restriction
        """
        return cls(datetime)

    @classproperty
    def DATE(cls):
        """
        Shortcut for a datetime.date restriction.
        :rtype: Restriction
        """
        return cls(date)

    @classproperty
    def NULL_DATETIME(cls):
        """
        Shortcut for a nullable datetime.datetime restriction.
        :rtype: Restriction
        """
        return cls(datetime, type(None))

    @classproperty
    def NULL_DATE(cls):
        """
        Shortcut for a nullable datetime.date restriction.
        :rtype: Restriction
        """
        return cls(date, type(None))

    @classproperty
    def BOOL_INT(cls):
        """
        Unlike the rest of the helpers, this is not a type restriction
            but a value restriction that represents bool as an integer of values 0 and 1.
        :rtype: Restriction
        """
        return cls(0, 1)

    @classproperty
    def LONG_INT(cls):
        """
        Shortcut for a long int restriction.
        :rtype: Restriction
        """
        if sys.version_info.major == 3:
            return cls(int, newint.newint, long)
        return cls(int, newint, long)

    @classproperty
    def NULL_LONG_INT(cls):
        """
        Shortcut for a nullable long int restriction.
        :rtype: Restriction
        """
        if sys.version_info.major == 3:
            return cls(int, newint.newint, long, type(None))
        return cls(int, newint, long, type(None))



"""
Property decorators useful for organizing code in a DO.
:date_created: 2020-07-10
"""

from builtins import object
import inspect


class classproperty(object):
    """
    This is a decorator.
    https://stackoverflow.com/questions/3203286/how-to-create-a-read-only-class-property-in-python
    """

    def __init__(self, f):
        """
        f is a method in a class that should be a property. This function will be able to access the attribute from
        class-level. Instances are not required, but the attribute value in the instance is preferred over compile-time.
        :param f:
        :rtype: classmethod
        """
        self.f = f

    def __get__(self, instance, owner):
        return self.f(instance if instance else owner)


def cached_property(original_property):
    """
    Adds object caching capability to getter property. Value returned by property 'fn' is cached in '_fn' attr. This is
    useful when property is computationally expensive to calculate. By caching the computed value, repeated calls to the
    property become more efficient.

    Example:

    class A(object):
        @cached_property
        def ok(self):
            return 'ok'

    Assume that a call to 'ok' is expensive. The return value of ok will be object level cached the first time it is
    called. Subsequent calls will avoid computation. The trade-off is additional memory consumption due to storage of
    the cached value.

    :param original_property: method intended to be used as property.
    :type original_property: method
    :return: property is object level return value caching.
    :rtype: property
    """
    attr = '_%s' % original_property.__name__

    def worker(self):
        """
        Proxy property that performs object level caching.
        """
        if not hasattr(self, attr):
            setattr(self, attr, original_property(self))
        return getattr(self, attr)

    setattr(worker, '_cached_', True)
    worker.__name__ = original_property.__name__
    worker.__doc__ = original_property.__doc__
    return property(fget=worker)


def is_cached_property(_cached_property):
    """
    Check if a class's property is wrapped by `cached_property`.
    :param _cached_property: The property in question.
    :rtype: bool
    :see: cached_property
    """
    # Get the `fget` method from the property.
    fget = getattr(_cached_property, 'fget', None)
    if fget:
        return getattr(fget, '_cached_', False)
    else:
        return False


def is_classmethod(cls, method):
    """
    Check if a method is a classmethod. Supports Python 2/3.
    :ref: https://stackoverflow.com/questions/19227724/check-if-a-function-uses-classmethod
    :param cls: Class
    :type cls: type
    :param method: Name of the method to check
    :type method: callable
    :rtype: bool
    """
    return inspect.ismethod(method) and hasattr(method, '__self__') and getattr(method, '__self__') is cls


def is_property(attribute):
    """
    Check if a class attribute is of type property.
    :ref: https://stackoverflow.com/questions/17735520/determine-if-given-class-attribute-is-a-property-or-not-python-object
    :param attribute: The class's attribute that will be checked.
    :rtype: bool
    """
    return isinstance(attribute, property)



"""
:date_created: 2019-08-01
"""

import json

from do_py.utils.json_encoder import MyJSONEncoder


class RestrictedDictMixin(dict):
    """
    Mixin for DataObjects that drops support for methods that allow the structure to be mutable.
    """

    def _unsupported(self, op):
        raise TypeError('%s does not support %s' % (self.__class__.__name__, op))

    def __delitem__(self, k):
        """
        deletion is not allowed in RestrictedDict
        """
        self._unsupported('delete')

    def clear(self):
        """
        clear is not allowed in RestrictedDict
        """
        self._unsupported('clear')

    def pop(self, k, *args):
        """
        pop is not allowed in RestrictedDict
        """
        self._unsupported('pop')

    def popitem(self):
        """
        popitem is not allowed in RestrictedDict
        """
        self._unsupported('popitem')

    def update(self, e=None, **kwargs):
        """
        Equivalent to dict.update(), but it was needed to call RestrictedDict.__setitem__() instead of dict.__setitem__
        """
        self._unsupported('update')

    def __repr__(self):
        return json.dumps(self, cls=MyJSONEncoder)

    def __str__(self):
        return '%s%s' % (self.__class__.__name__, super(RestrictedDictMixin, self).__str__())

    def __setitem__(self, item, value):
        """
        This assigns a value to item in the key and attribute namespaces.
        :param item:
        :param value:
        """
        super(RestrictedDictMixin, self).__setitem__(item, value)

    def __setattr__(self, key, value):
        """
        This assigns a value to key in the attribute namespace. This value will not undergo data validation, unless the
        key exists in the key namespace.
        Note: Value in key namespace is preferred over attribute namespace.
        :param key: dictionary key
        :param value: value to store in dict or instance
        """
        if key in self:
            self[key] = value
        else:
            super(RestrictedDictMixin, self).__setattr__(key, value)

    def __getattr__(self, item):
        """
        Only called when item is not found in the attribute namespace, meaning that we only need to check
        if item exists in the key space.
        :param item: dictionary key
        :return: attribute selected by item key
        :raises AttributeError: KeyErrors are reraised as attribute errors to match attribute getter.
        """
        try:
            return self[item]
        except KeyError:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, item))

    def __dir__(self):
        dict_dir = dir(dict)
        for x in ['clear', 'pop', 'popitem', 'update', '__delitem__']:
            dict_dir.remove(x)
        return dict_dir



"""
Dynamic restrictions for a DataObject.
:date_created: 2020-07-10
:author: Gian Brazzini
"""
from do_py import DataObject, R
from do_py.abc import ABCRestrictionMeta
from do_py.data_object.restriction import ManagedRestrictions, _ListValueRestriction
from do_py.utils.properties import cached_property


class DynamicRestrictions(ManagedRestrictions):
    """
    A dict that contains the dynamic restrictions where the key is the independent
    key's value and the value is the dependent key's restriction.
    """
    _restriction = R()

    def manage(self):
        """
        Validate that the list is not empty and that the format is correct.
        """
        assert self.data, 'Missing dynamic restrictions, dict cannot be empty.'
        for value in self.data.values():
            assert DataObject in value.mro(), 'Dynamic restrictions depend each value being a DataObject.'


class DynamicClassGenerator(DataObject):
    """
    Generate a class that is inheritable by a DO that requires dynamic restrictions.
    :restriction independent_key: The key that the dynamic restriction depends on.
    :restriction dependent_key: The key where the restriction is dynamic.
    :restriction dynamic_restrictions: See `DynamicRestrictions` ManagedList.
    """
    _restrictions = {
        'independent_key': R.STR,
        'dependent_key': R.STR,
        'dynamic_restrictions': DynamicRestrictions()
        }

    @property
    def update_fn_name(self):
        """
        :rtype: str
        """
        return '_update_%s_restriction' % self.dependent_key

    @cached_property
    def dynamic_class(self):
        """
        Create a class dynamically which inherits from DataObject, but also contains the required attributes to
        create dynamic restrictions for the dependent key.
        :rtype: type(DataObject)
        """

        def update_restriction(instance_self):
            """
            Update the dynamic restriction. This function will live inside the instance of the
            class under the name declared in `update_fn_name`.
            """
            restriction = self.dynamic_restrictions[instance_self[self.independent_key]]
            # Update the restriction.
            instance_self._restrictions[self.dependent_key] = restriction
            # Validate the data with the new restriction.
            instance_self._restrictions[self.dependent_key](instance_self[self.dependent_key])
            # Set the dynamic in the instance so that we can verify the data.
            super(self.dynamic_class, instance_self).__setitem__(self.dependent_key, instance_self[self.dependent_key])

        attributes = {
            '_is_abstract_': True,
            '__module__': self.__class__.__module__,
            self.update_fn_name: update_restriction,
            }

        # Create the class using ABCRestrictionMeta as the metaclass and inheriting from DataObject.
        return ABCRestrictionMeta('dynamic_%s_mixin' % self.dependent_key, (DataObject,), attributes)

    @cached_property
    def init_method(self):
        """
        The init method needs to call the update restriction method after init is done.
        :rtype: types.Callable
        """

        def __init__(instance_self, data, **init_kwargs):
            super(self.dynamic_class, instance_self).__init__(data, **init_kwargs)
            getattr(instance_self, self.update_fn_name)()

        return __init__

    @cached_property
    def setitem_method(self):
        """
        Generate the set item method which overrides the parent's. When either the dependent or independent keys
        are updated, we must run the update restriction method.
        :rtype: types.Callable
        """

        def __setitem__(instance_self, item_key, item_value):
            response = super(self.dynamic_class, instance_self).__setitem__(item_key, item_value)
            if item_key in [self.independent_key, self.dependent_key]:
                getattr(instance_self, self.update_fn_name)()
            return response

        return __setitem__

    @cached_property
    def compile_classmethod(self):
        """
        Generate the classmethod that will be used for compilation.
        The following validations are made:
        1) The independent and dependent keys must exist in the restrictions.
        2) The independent restriction is a list value restriction.
        3) All values for the independent restrictions are accounted for in the dynamic restrictions.
        :rtype: classmethod
        """

        def __compile__(cls):
            super(self.dynamic_class, cls).__compile__()

            # Validate that the independent and dependent keys must exist in the restrictions.
            assert self.independent_key in cls._restrictions, \
                '%s.%s required in restrictions for dynamic restrictions.' % (cls.__name__, self.independent_key)
            assert self.dependent_key in cls._restrictions, \
                '%s.%s required in restrictions for dynamic restrictions.' % (cls.__name__, self.dependent_key)

            # Validate that the independent restriction is a list value restriction.
            independent_key_restriction = cls._restrictions[self.independent_key]
            assert isinstance(independent_key_restriction, _ListValueRestriction), \
                'The independent key must be of type `_ListValueRestriction`.'

            # Validate that all values for the independent restrictions are accounted for in the dynamic restrictions.
            assert len(independent_key_restriction[0]) == len(self.dynamic_restrictions), \
                'All restrictions must be accounted for when creating dynamic restrictions.'
            for value in independent_key_restriction[0]:
                assert value in self.dynamic_restrictions, 'Missing dynamic restriction for value "%s".' % value

        return classmethod(__compile__)

    @cached_property
    def mixin(self):
        """
        The finished product for the generated class mixin. It had to be done in two
            properties due to the method dependency noted below.
        :rtype: type(DataObject)
        """
        # Methods that require the use of super must be attached after instantiation.
        self.dynamic_class.__init__ = self.init_method
        self.dynamic_class.__setitem__ = self.setitem_method
        self.dynamic_class.__compile__ = self.compile_classmethod
        return self.dynamic_class


def dynamic_restriction_mixin(independent_key, dependent_key, **kwargs):
    """
    Create a mixin class that a DO class can inherit to enable dynamic restrictions.
    See `DynamicClassGenerator` for further instruction.
    :param independent_key: The key that the dynamic restriction depends on.
    :param dependent_key: The key where the restriction is dynamic.
    :param kwargs: The dynamic restrictions.
    :rtype: type(DataObject)
    """
    return DynamicClassGenerator({
        'independent_key': independent_key,
        'dependent_key': dependent_key,
        'dynamic_restrictions': kwargs,
        }).mixin



"""
:date_created: 2020-06-28
"""
from do_py.common import R
from do_py.data_object.restriction import ManagedRestrictions
from do_py.exceptions import RestrictionError


class ManagedList(ManagedRestrictions):
    """
    Use this when you need a restriction for a list of DataObject's.
    """
    _restriction = R(list, type(None))

    def __init__(self, obj_cls, nullable=False):
        """
        :param obj_cls: The DO to check each value in the list against.
        :type obj_cls: DataObject
        :param nullable: Valid values are a list of Do's or a NoneType.
        :type nullable: bool
        """
        super(ManagedList, self).__init__()
        self.obj_cls = obj_cls
        self.nullable = nullable

    def manage(self):
        if self.data is not None:
            items = []
            for item in self.data:
                items.append(item if type(item) == self.obj_cls else self.obj_cls(item))
            self.data = items
        else:
            if not self.nullable:
                raise RestrictionError.bad_data(self.data, self._restriction.allowed)


# TODO: Unit tests
class OrderedManagedList(ManagedList):

    def __init__(self, obj_cls, nullable=False, key=None, reverse=False):
        """
        :param obj_cls: DataObject class reference to wrap each object in list.
        :type nullable: bool
        :type key: function
        :type reverse: bool
        """
        self.key = key
        self.reverse = reverse
        super(OrderedManagedList, self).__init__(obj_cls, nullable=nullable)

    def manage(self):
        """
        Sort the data list after ManagedList does its work.
        """
        super(OrderedManagedList, self).manage()
        self.data = sorted(self.data, key=self.key, reverse=self.reverse)



"""
Custom Abstract Base Classes to implement Restrictions. Developed for DataObject concept.
:date_created: 2018-12-05
"""

from builtins import object

from .constants import ConstABCR
from .messages import SystemMessages
from .utils import already_declared, compare_cls
from ..utils import classproperty


class ABCRestrictionMeta(type):
    """
    TODO
    """
    _abc_classes = set()
    _unique_attrs = dict()

    @classproperty
    def abc_classes(cls):
        """
        To support parametrizing unit tests.
        :return:
        """
        return {k: getattr(k, ConstABCR.required, ()) for k in cls._abc_classes}

    def __new__(mcs, cls_name, parents, namespace):
        """

        :param cls_name:
        :param parents:
        :param namespace:
        :return:
        """
        assert '__module__' in namespace, SystemMessages.REQUIRED_FOR % ('__module__', 'ABCRestrictionsMeta')

        def nested_new(fn_new):
            def this_new(this_cls, *args, **kwargs):
                # Default abstract class check
                if getattr(this_cls, ConstABCR.state, None) != ConstABCR.leaf:
                    compare_cls((this_cls.__module__, this_cls.__name__), (namespace['__module__'], cls_name))

                if fn_new:
                    # Run explicitly defined __new__
                    assert callable(fn_new) or hasattr(fn_new, '__func__'), \
                        'Failed to decipher how to run nested __new__ for %s' % this_cls
                    if callable(fn_new):
                        return fn_new(this_cls, *args, **kwargs)
                    elif hasattr(fn_new, '__func__'):
                        return fn_new.__func__(this_cls, *args, **kwargs)
                else:
                    # Run default __new__
                    for p_ in parents:
                        if hasattr(p_, '__new__'):
                            p_.__new__(this_cls, *args, **kwargs)  # Unsure if trashing returned instance is a problem
                    return dict.__new__(this_cls, *args, **kwargs)

            return this_new

        namespace[ConstABCR.new] = nested_new(namespace.get(ConstABCR.new))

        if ConstABCR.required in namespace:
            # Root-type and Node-type classes define namespace requirements for its children.
            assert namespace.get(ConstABCR.state, ConstABCR.root) in ConstABCR.restrictions_allowed, \
                'Invalid class type=%s' % namespace.get(ConstABCR.state, ConstABCR.root)
            namespace[ConstABCR.state] = namespace.get(ConstABCR.state, ConstABCR.root)

            cls = type.__new__(mcs, cls_name, parents, namespace)
            mcs._abc_classes.add(cls)
        elif namespace.get(ConstABCR.is_abstract):
            # Node-style classes pass on requirements from roots to children of nodes. This class cannot be initialized.
            # Extra functionality can be added by explicitly declaring __new__
            namespace[ConstABCR.state] = ConstABCR.node
            cls = type.__new__(mcs, cls_name, parents, namespace)
        else:
            # Leaf-type classes implement requirements defined from Root and Node-type classes
            namespace[ConstABCR.state] = ConstABCR.leaf

            roots = []
            nodes = []
            leaves = []
            all_parents = {k for k in list(parents) + sum([p.mro() for p in parents], [])}
            for c in all_parents:
                if getattr(c, ConstABCR.state, None) == ConstABCR.root:
                    roots.append(c)
                elif getattr(c, ConstABCR.state, None) == ConstABCR.node:
                    nodes.append(c)
                elif getattr(c, ConstABCR.state, None) == ConstABCR.leaf:
                    leaves.append(c)

            # Validate that all required attributes from this class's roots are fulfilled in one of the following:
            #   1. This leaf's namespace
            #   2. One of its leaf-style parents' namespaces
            #   3. One of its node-style parents' namespaces
            required_attrs = set(sum([getattr(p, ConstABCR.required, ()) for p in roots + nodes], ()))
            for attr in required_attrs:
                # Check that the required attribute is defined in this class or a leaf that is a parent of this class.
                assert any([hasattr(p, attr) for p in leaves + nodes]) or attr in namespace, \
                    SystemMessages.REQUIRED_FOR % (attr, cls_name)

            # Validate that the value given to a unique attribute is unique system-wide for that attribute.
            unique_attrs = set(sum([getattr(p, ConstABCR.unique, ()) for p in roots + nodes], ()))
            for attr in unique_attrs:
                for leaf in mcs._unique_attrs.get(attr, []):
                    if attr not in namespace:
                        # This means the unique attr should be fulfilled in a leaf or node parent. Uniqueness check was
                        # already done at that level, so we can skip.
                        continue

                    if (namespace['__module__'], cls_name) == (leaf.__module__, leaf.__name__):
                        # Re-declaring the class... This supports iPython.
                        mcs._unique_attrs[attr].remove(leaf)
                        continue

                    assert namespace[attr] != getattr(leaf, attr), \
                        'Unique value "%s" has already been declared in class %s' % (namespace[attr], leaf.__name__)

            cls = type.__new__(mcs, cls_name, parents, namespace)

            # Register this class against the unique_attrs it fulfills for other leaves to be able to value check.
            for attr in unique_attrs:
                if attr not in mcs._unique_attrs:
                    mcs._unique_attrs[attr] = []
                assert hasattr(cls, attr), SystemMessages.REQUIRED_FOR % (attr, cls_name)
                mcs._unique_attrs[attr].append(cls)

            # Run optional compile-time validation function
            if hasattr(cls, '__compile__'):
                cls.__compile__()

        return cls


class ABCRestrictions(object):
    """
    ABCRestrictionMeta implementation via decorators.
    """

    @classmethod
    def require(cls, *required_attrs, **kwargs):
        """
        Decorator to attach requirements on the attribute namespace and declare the class as an abstract class.
        :param required_attrs:
        :keyword unique: subset of required_attrs that are required to have unique values system-wide
        :return: decorated class
        :raises: NotImplementedError if __new__ or __metaclass__ already defined.
        :raises: AssertionError if class reference already registered. (namespace clash)
        :raises: AssertionError if class is already declared as a Leaf-type class.
        :raises: AssertionError if a required attribute is already declared in a parent.
        """

        def worker(cls_ref):
            """

            :param cls_ref:
            :return:
            """
            assert cls_ref not in ABCRestrictionMeta.abc_classes, SystemMessages.CLASS_ALREADY_DEFINED
            assert getattr(cls_ref, ConstABCR.state, ConstABCR.root) in ConstABCR.restrictions_allowed, \
                SystemMessages.RESTRICTIONS_ALLOWED

            # Manipulate the namespace of the declared class so that metaclass handles it properly
            namespace = dict(cls_ref.__dict__)
            namespace.pop('__dict__', None)
            namespace.pop('__weakref__', None)

            # Check if required attributes passed in are new
            r = already_declared(cls_ref, ConstABCR.required, required_attrs)
            assert not r, SystemMessages.ATTRIBUTE_ALREADY_DECLARED % ('Required', r,
                                                                       "%s's parents" % cls_ref.__name__)
            namespace[ConstABCR.required] = required_attrs + getattr(cls_ref, ConstABCR.required, ())

            # Check if uniqueness requirements are declared for some attributes
            if kwargs.get('unique'):
                # Validate unique passed in properly
                unique = kwargs['unique']
                assert type(unique) in [list, tuple], 'Invalid type "%s" for parameter "unique"' % unique
                assert all([e in required_attrs for e in unique]), 'Unique attributes must be in required attributes!'

                # Validate unique attributes are new across all abc_classes
                for abc_cls in ABCRestrictionMeta.abc_classes:
                    u = already_declared(abc_cls, ConstABCR.unique, unique)
                    assert not u, SystemMessages.ATTRIBUTE_ALREADY_DECLARED % ('Unique', u, abc_cls.__name__)
                namespace[ConstABCR.unique] = tuple(unique) + getattr(cls_ref, ConstABCR.unique, ())
            return ABCRestrictionMeta(cls_ref.__name__, cls_ref.__bases__, namespace)

        return worker



"""
Utils to support base_model modules.
:date_created: 2018-12-05
:author: Tim Davis
"""

from do_py.abc.messages import SystemMessages


def compare_cls(cls_name, other):
    """
    Util to adhere to DRY concept for comparing class names. This is used to make sure abstract classes are
    not instantiated.
    :param cls_name:
    :param other:
    :raises: NotImplementedError
    """
    if cls_name == other:
        raise NotImplementedError(SystemMessages.INSTANTIATION_ERROR)


def already_declared(_cls_ref, attr_name, attrs):
    if getattr(_cls_ref, attr_name, ()):
        for attr in attrs:
            if attr in getattr(_cls_ref, attr_name):
                return attr



"""
Constants to support base_model modules.
:date_created: 2018-12-05
"""

from builtins import object


class ConstABCR(object):
    """
    Constants supporting ABCRestrictions
    """
    required = '_required_'
    unique = '_unique_'
    metaclass = '__metaclass__'
    is_abstract = '_is_abstract_'
    new = '__new__'
    state = '_state_'
    # Class types
    root = 'root'
    node = 'node'
    leaf = 'leaf'
    restrictions_allowed = [root, node]



"""
:date_created: 2019-08-01
"""
import json

from datetime import date, datetime


class MyJSONEncoder(json.JSONEncoder):
    """
    Custom JSON encoder used for adding encoding support to datetime/date.
    """

    def default(self, obj):
        """
        :param obj: Object being encoded.
        :return: Encoded format.
        """
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, date):
            return obj.isoformat()
        return super(MyJSONEncoder, self).default(obj)



"""
System messages to directly support error messages in base_model modules.
:date_created: 2018-12-05
:author: Tim Davis
"""

from builtins import object


class SystemMessages(object):
    REQUIRED_FOR = '%s is required for %s!'
    GENERIC = 'Unknown error.'
    INSTANTIATION_ERROR = 'Base class may not be instantiated'
    RESTRICTIONS_ALLOWED = 'Requirements can only be assigned to Root and Node-type classes!'
    CLASS_ALREADY_DEFINED = 'Class already defined in ABCRestrictions!'
    ATTRIBUTE_ALREADY_DECLARED = '%s attribute "%s" has already been declared in %s!'



"""
Data Object Restrictions.
:date_created: 2019-05-13
"""

import copy
import sys
from abc import ABCMeta, abstractmethod, abstractproperty

from builtins import object
from datetime import date, datetime

from future.moves import builtins
from future.types import newint, newlist, newstr
from future.utils import with_metaclass
from past.builtins import long, unicode

from ..abc import ABCRestrictionMeta
from ..exceptions import RestrictionError
from ..utils import classproperty


class AbstractRestriction(tuple):
    """
    Manage restriction syntax. Pull allowed and default values for the tuple syntax.

    E.g.
    For R.INT, allowed = [int] and default = None

    Syntax:
    Restrictions are typically declared as tuples. The first item in the restriction is allowed. The second item is
    the default value. See specific restriction for more details.

    allowed:
    Data restrictions are captured as allowed. See specific restriction for more details.

    default:
    Every restriction provides a default value. If not explicitly specified, it is implicitly set to None.
    """
    _default = None
    _allowed = None

    def __init__(self, *args, **kwargs):
        """
        default and allowed are provided as tuple.
        """
        if len(self) > 1:
            self._default = self[1]

        self._allowed = self[0]

    def __call__(self, data, **kwargs):
        """
        Execute data validation. See specific restriction for more details.
        """

    def with_default(self, default):
        """
        w: short for `with_default`
        Add/update the default of an existing restriction.
        :param default: The new default for this cls.
        :return: cls.__class__
        """
        return self.__class__(self._allowed, default)

    @property
    def allowed(self):
        return self._allowed

    @property
    def default(self):
        return self._default

    # NOTE:
    # Opportunity 1: ES restriction building capability can be triggered from es_ables.
    # Opportunity 2: The encoder system can be abstracted to support multiple encoding needs (including custom).
    # Possibility to share restrictions between python and JS.
    # Techs similar to ES may use different encoding systems.
    #
    # E.g. default json package in python supports custom encoding.
    #   json.loads(x, cls=CustomEncoder)
    #
    # See https://docs.python.org/2/library/json.html for more details on this example. ComplexEncoder is custom
    # encoder.
    #
    # import json
    # >>> class ComplexEncoder(json.JSONEncoder):
    # ...     def default(self, obj):
    # ...         if isinstance(obj, complex):
    # ...             return [obj.real, obj.imag]
    # ...         # Let the base class default method raise the TypeError
    # ...         return json.JSONEncoder.default(self, obj)
    # ...
    # >>> json.dumps(2 + 1j, cls=ComplexEncoder)
    # '[2.0, 1.0]'
    # >>> ComplexEncoder().encode(2 + 1j)
    # '[2.0, 1.0]'
    # >>> list(ComplexEncoder().iterencode(2 + 1j))
    # ['[', '2.0', ', ', '1.0', ']']
    @property
    def es_restrictions(self):
        raise NotImplementedError('ES restrictions not defined.')

    def __copy__(self):
        return tuple(self)

    def __deepcopy__(self, memodict={}):
        memodict[id(self)] = self.__copy__()
        _r = copy.deepcopy(self.__copy__())
        return self.__class__(_r[0], default=_r[1])


def is_immutable(obj):
    return any(isinstance(obj, x) for x in [bool, int, float, tuple, str, frozenset, type(None)])


class SingletonRestriction(AbstractRestriction):
    """
    This is an interface for Restriction type to use singleton structure. The objective is to use pre-defined
    restrictions and reduce the memory footprint of DataObject declarations.
    """
    _cache = {}

    def __new__(cls, restriction_tuple):
        """
        This will hash and stash this restriction to allow a singleton instance to be used throughout the entire system
        per system instance.
        :param restriction_tuple: (allowed, default)
        :type restriction_tuple: tuple
        :rtype: SingletonRestriction
        """
        try:
            rt1 = None
            if is_immutable(restriction_tuple[1]):
                rt1 = restriction_tuple[1]
            elif type(restriction_tuple[1]) in [datetime, date]:
                # date and datetime instances are mutable. They are not iterable, hence not hashable.
                rt1 = restriction_tuple[1].isoformat()
            else:
                rt1 = frozenset(restriction_tuple[1])

            # cls.__name__ supports restriction inheritance, i.e. _NullableDataObjectRestriction
            if type(restriction_tuple[0]) is ABCRestrictionMeta:
                hashable = (cls.__name__, restriction_tuple[0])
            else:
                hashable = (cls.__name__, frozenset(restriction_tuple[0]), rt1)
        except TypeError:
            raise RestrictionError.from_unhashable(restriction_tuple[0], restriction_tuple[1])

        if hashable in cls._cache:
            return cls._cache[hashable]
        else:
            cls._cache[hashable] = super(SingletonRestriction, cls).__new__(cls, restriction_tuple)
            return cls._cache[hashable]

    @property
    def es_restrictions(self):
        raise NotImplementedError('ES restrictions not defined.')


class _ListTypeRestriction(SingletonRestriction):
    """
    Manage restriction of syntax ([type], None)

    Syntax:
    This uses standard tuple syntax.

    allowed:
    allowed is declared as a list of types.

    default:
    None if not specified.

    Validation:
    The type of supplied data must be an item in the allowed list.

    E.g.:

    class A(DataObject):
        _restrictions = {
            'x': R.INT
            }

    Here, x can be numbers like 1, 1.0, -3, etc.
    Good for some data like pulsechecks sent (x >= 0).
    """

    def __new__(cls, allowed, default=None, **kwargs):
        return super(_ListTypeRestriction, cls).__new__(cls, (allowed, default))

    def __init__(self, *args, **kwargs):
        super(_ListTypeRestriction, self).__init__()
        if len([e for e in self._allowed if type(e) is ABCRestrictionMeta]) != 0:
            raise RestrictionError.from_dataobj_in_rstr_list(self._allowed)

        if not all([isinstance(r, type) for r in self._allowed]):
            raise RestrictionError.from_mixed_value_and_type(self._allowed)

    def __call__(self, data, **kwargs):
        if type(data) not in self._allowed:
            raise RestrictionError.bad_data(type(data), self._allowed)
        return data

    @property
    def es_restrictions(self):
        _x = set()
        for e in self.allowed:
            if e not in [None, type(None)]:
                for i in ESEncoder.default(e).items():
                    _x.add(i)
        if len(_x) == 1:
            return dict([_x.pop()])
        raise RestrictionError('Ambiguous ES restricitons.')


class _ListValueRestriction(SingletonRestriction):
    """
    Manage restriction of syntax ([values], None)

    Syntax:
    This uses standard tuple syntax.

    allowed:
    allowed is declared as a list of values.

    default:
    None if not specified.

    Validation:
    Supplied data must be an item in the allowed list.

    E.g.:

    class A(DataObject):
        _restrictions = {
            'x': R.INT
            }

    Here, x can be numbers like 1, 2 and 3. However, numbers like 1.0, -3, etc are not allowed. Typically, this is
    more restrictive than _ListTypeRestriction.

    Good choice for some data such as Review ratings (1 <= ratings <= 5).
    """

    def __new__(cls, allowed, default=None, **kwargs):
        return super(_ListValueRestriction, cls).__new__(cls, (allowed, default))

    def __call__(self, data, **kwargs):
        if data not in self._allowed:
            raise RestrictionError.bad_data(data, self._allowed)
        return data

    @property
    def es_restrictions(self):
        return ESEncoder.default('keyword')


class _ListNoRestriction(SingletonRestriction):
    """
    Manage restriction of syntax R()

    Syntax:
    Empty list for allowed.

    allowed:
    allowed is declared as an empty list.

    default:
    None if not specified.

    Validation:
    Any data is allowed. This is the most flexible restriction.

    E.g.:

    class A(DataObject):
        _restrictions = {
            'x': R()
            }

    Here, x can be primitive data types like numbers, strings, boolean, etc.
    It can also be more complex data such as dict, list, set, classes, instances, etc.

    Good choice for data where no validation is necessary.
    """

    def __new__(cls, allowed, default=None, **kwargs):
        return super(_ListNoRestriction, cls).__new__(cls, (allowed, default))

    def __init__(self, *args, **kwargs):
        super(_ListNoRestriction, self).__init__()
        assert isinstance(self._allowed, list) and len(self._allowed) == 0, 'Complain'

    def __call__(self, data, **kwargs):
        return data

    @property
    def es_restrictions(self):
        return None


class _MgdRestRestriction(AbstractRestriction):
    """
    Manages restriction of type ManagedRestrictions.

    Syntax:
    Specified as an instance of ManagedRestrictions class. See example below.

    allowed:
    allowed is an instance of ManagedRestrictions class.

    default:
    None by default. User can take-over default management in their implementation of ManagedRestrictions.

    Validation:
    Validation is defined by user in ManagedRestrictions class, more specifically, manage method.

    E.g.:

    class ManagedX(ManagedRestrictions):
        _restriction = R.INT

        def manage(self):
            assert self.data % 2, 'Even numbers not allowed'

    class A(DataObject):
        _restrictions = {
            'x': ManagedX()
            }

    In this example, we allow odd numbers for x. Even numbers are not allowed.
    Good choice for custom data validation.

    For more information on ManagedRestrictions, see ManagedRestrictions class.
    """

    def __new__(cls, allowed, default=None, **kwargs):
        return super(_MgdRestRestriction, cls).__new__(cls, (allowed, default))

    def __call__(self, data, **kwargs):
        return self._allowed(data)

    @property
    def es_restrictions(self):
        # NOTE: _allowed is of ManagedRestrictions type. _restriction is of AbstractRestriction type
        return self._allowed._restriction.es_restrictions


class _NullableDataObjectRestriction(SingletonRestriction):
    """
    Manage restriction of type [DataObject, type(None)].

    Syntax:
    Specified as a list containing a class reference to a DO and type(None). See example below.

    allowed:
    allowed is a list containing a class reference to a DO and type(None).

    default:
    None by default.

    Validation:
    Validation is performed by the DataObject. None is allowed.

    E.g.:

    class B(DataObject):
        _restrictions = {
            'x': R.INT
            }

    class A(DataObject):
        _restrictions = {
            'b': B
            }

    In this example, the value of key 'b' in A._restrictions will be validated by class B, which is also a DataObject.
    A value of None is also allowed for 'b'.
    """

    def __new__(cls, allowed, default=None, **kwargs):
        return super(_NullableDataObjectRestriction, cls).__new__(cls, (allowed, default))

    def __call__(self, data, strict=True, **kwargs):
        if data is None:
            return data
        elif strict and not isinstance(data, dict):
            raise RestrictionError.bad_data(data, self._allowed)
        return self._allowed(data=data, strict=strict)

    @property
    def dataobj(self):
        return self.allowed

    @property
    def es_restrictions(self):
        if hasattr(self.dataobj, 'es_restrictions'):
            return self.dataobj.es_restrictions
        else:
            return {
                'properties': {
                    k: v.es_restrictions for k, v in self.dataobj._restrictions.items()
                    }
                }


class _DataObjectRestriction(_NullableDataObjectRestriction):
    """
    Manages restriction of type DataObject.

    Syntax:
    Specified as a class refence to a DO. See example below.

    allowed:
    allowed is a class reference to a DO.

    default:
    None by default.

    Validation:
    Validation is performed by the DataObject.

    E.g.:

    class B(DataObject):
        _restrictions = {
            'x': R.INT
            }

    class A(DataObject):
        _restrictions = {
            'b': B
            }

    In this example, the value of key 'b' in A._restrictions will be validated by class B, which is also a DataObject.
    Complex data requires data nesting. This typically requires nested DOs, which is supported by this restriction.
    """

    def __new__(cls, allowed, default=None, **kwargs):
        default = allowed(strict=False)
        return super(_DataObjectRestriction, cls).__new__(cls, allowed, default=default)

    def __call__(self, data, strict=True, **kwargs):
        if type(data) is self._allowed:
            return data
        return self._allowed(data=data, strict=strict)


class Restriction(object):
    """
    Restriction factory which manages restriction delegation.
    """

    def __new__(cls, allowed, default=None, **kwargs):
        """
        :param allowed: Allowed per restrictions.
        :type allowed: list or ManagedRestrictions or ABCRestrictionMeta
        :param default: Default value.
        """
        if type(default) is type:
            raise RestrictionError.from_invalid_default_value(default)
        if isinstance(allowed, ManagedRestrictions):
            return _MgdRestRestriction(allowed, default=allowed.default, **kwargs)
        elif type(allowed) in [list, newlist]:
            if len(allowed) == 0:
                return _ListNoRestriction(allowed, default=default, **kwargs)
            elif len(allowed) == 2 and ABCRestrictionMeta in [type(e) for e in allowed]:
                if type(None) not in allowed:
                    raise RestrictionError.from_unsupported_dataobj_in_rstr_list(allowed)
                return _NullableDataObjectRestriction(allowed[1 - allowed.index(type(None))], default=default, **kwargs)
            elif any([isinstance(r, type) for r in allowed]):
                return _ListTypeRestriction(allowed, default=default, **kwargs)
            else:
                return _ListValueRestriction(allowed, default=default, **kwargs)
        elif type(allowed) is ABCRestrictionMeta:
            return _DataObjectRestriction(allowed, default=default, **kwargs)
        else:
            raise RestrictionError.from_unsupported(allowed)

    def __init__(cls, allowed, default=None, **kwargs):
        """
        Dummy init to support classmethod(s)
        """

    @classmethod
    def legacy(cls, declaration):
        """
        Support of legacy tuple syntax.

        Legacy syntax:
        In current usage of DataObjects, restrictions declarations do not use the specific restriction classes. E.g.,
        ([int, float], None)

        Restriction class syntax:
        The example above can be also be declared using restriction class, i.e.,
        _ListTypeRestriction([int, float], default=None)

        Vision:
        2019-06-06:
        As of this day, the plan is to expose the restrictions classes using simplied syntax. Example,

            R(int, default=1)

        Also, there is a plan to collect the most common restrictions into a STL (standard template library). Example,

            In STL,

            R_int = R([int, float])

            In user code,

            class A(DataObject):
                _restrictions = {
                    'x': R.INT
                    }

        :param declaration: Legacy tuple syntax
        :type declaration: tuple or list or ManagedRestrictions or ABCRestrictionMeta
        """
        # Python 3 format
        if isinstance(declaration, AbstractRestriction):
            # NOTE: This is already a restriction instance, so nothing to do here.
            return declaration
        elif isinstance(declaration, ABCRestrictionMeta) or isinstance(declaration, ManagedRestrictions):
            return cls(declaration)
        elif type(declaration) is list:
            if any([isinstance(r, type) for r in declaration]):
                raise RestrictionError.from_unsupported(declaration)
            return cls(declaration)
        elif type(declaration) is tuple:
            # Legacy format
            raise RestrictionError.from_unsupported(declaration)
        else:
            return cls(declaration)


class ManagedRestrictions(object, with_metaclass(ABCMeta)):
    """
    Useful for managing complex data validations for restrictions in DataObject. E.g.
    1. Regular expression validation such as review URL matches expected template.
    2. Phone number validation and standardization such that number is in ISO format and is SMS enabled.

    DataObject understands that restrictions can be mapped to classes inherited from ManagedRestrictions. When
    DataObject encounters a ManagedRestriction, a call to manage method is made. Managed restriction value is stored in
    data attribute.

    Example:
    A toy example to demonstrate usage. 'name' is a key for Contact and it is managed by Name class. Functionally, the
    Name class ensures that key 'name' is always in title case.

    class Name(ManagedRestrictions):
        _restriction = R.STR

        def manage(self):
            # validation
            assert self.data, 'Invalid name'
            # standardization
            self.data = self.data.title()

    class Contact(DataObject):
        _restrictions: {
            'name': Name(),
            }

    :attribute manage: users must implement validation/standardization logic in manage
    """
    data = None

    def __new__(cls, *args, **kwargs):
        cls._restriction = Restriction.legacy(cls._restriction)
        return super(ManagedRestrictions, cls).__new__(cls)

    @abstractproperty
    def _restriction(self):
        """
        Specify restriction for the key this class manages. Follow the same format as _restrictions in DataObject.
        """

    @abstractmethod
    def manage(self):
        """
        Data management logic is handled in this method. This can be used to standardize and/or validate data. Users of
        this class must implement this method. The value to manage can be found in self.data. It is best practice to
        raise exception if data fails validation.
        """

    @property
    def default(self):
        """
        Expose restriction default.
        """
        return self._restriction.default

    def __call__(self, value, strict=True):
        """
        Entry point for data management. DataObject will trigger a call to __call__ when it encounters a managed
        restriction.
        :param value: Data value that needs to be managed
        :param strict: Validation strictness.
        :return: Standardized data
        """
        data_copy = copy.deepcopy(self.data)
        try:
            self.data = value
            if strict:
                self.manage()
        except Exception:
            self.data = data_copy
            raise

        return self.data

    def __eq__(self, other):
        return self._restriction == other._restriction


# NOTE: ESR should be refactored to common.py, but OOS for now due to circular imports with ESEncoder
class ESR(object):
    """
    Constants class housing the different types of ES restrictions.
    """
    INT = {'type': 'integer'}
    FLOAT = {'type': 'float'}
    DATE = {'type': 'date'}
    BOOL = {'type': 'boolean'}
    STR = {'type': 'text'}
    KEYWORD = {'type': 'keyword'}

    @staticmethod
    def OBJECT(nested_data_object):
        """
        Object-type ElasticSearch restrictions. This restriction defines a nested object on ES mapping.
        :type nested_data_object: DataObject
        :rtype: dict
        """
        if not hasattr(nested_data_object, 'es_restrictions'):
            raise RestrictionError.bad_data(nested_data_object, 'DataObject')
        return {
            'type': 'object',
            'properties': nested_data_object.es_restrictions
            }

    @staticmethod
    def NESTED(nested_data_object):
        """
        Nested-type ElasticSearh restriction. This restriction defines a list of nested objects on ES mapping.
        :type nested_data_object: DataObject
        :rtype: dict
        """
        if not hasattr(nested_data_object, 'es_restrictions'):
            raise RestrictionError.bad_data(nested_data_object, 'DataObject')
        return {
            'type': 'nested',
            'properties': nested_data_object.es_restrictions
            }


class ESEncoder(object):
    """
    Convert restrictions into ES Document Property Map
    Ref: https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping-types.html
    """

    @classproperty
    def encoding(cls):
        """
        :rtype: dict
        """
        encoding = {
            int: ESR.INT,
            long: ESR.INT,
            builtins.int: ESR.INT,
            float: ESR.FLOAT,
            builtins.float: ESR.FLOAT,
            datetime: ESR.DATE,
            date: ESR.DATE,
            bool: ESR.BOOL,
            str: ESR.STR,
            builtins.str: ESR.STR,
            unicode: ESR.STR,
            'keyword': ESR.KEYWORD,
            # list: {},
            }
        if sys.version_info.major == 3:
            encoding.update({
                newint.newint: ESR.INT,
                newstr.newstr: ESR.STR
                })
        return encoding

    @classmethod
    def default(cls, obj):
        try:
            return cls.encoding[obj]
        except KeyError:
            raise RestrictionError('%s cannot be encoded!' % obj)



class DataObjectError(Exception):
    """
    Errors related to restriction and data keys.
    """

    @classmethod
    def from_unknown_key(cls, key, cls_ref):
        """
        Compile time error. Not allowed to pass a key to DO that was not declared in restrictions.
        """
        return cls("%s: Unexpected key '%s' in data." % (cls_ref.__name__, key))

    @classmethod
    def from_required_key(cls, key, cls_ref):
        """
        Run time error. A required key was absent in data.
        """
        return cls("%s: Key '%s' required in data." % (cls_ref.__name__, key))

    @classmethod
    def from_restriction_error(cls, key, cls_ref, restriction_error):
        """
        Run time error. An error captured from restriction layer is enhanced by adding class reference and key info.
        Finally rethrown.

        Movation:
        Catch -> Rethrow to add more information that would save debugging time.
        """
        return cls("%s.%s: %s" % (cls_ref.__name__, key, restriction_error))



"""
Setup script
"""
from __future__ import print_function

import json
import os
import re

import setuptools


dir_path = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(dir_path, 'README.md'), 'r') as fh:
    long_description = fh.read()

package = json.load(open(os.path.join(dir_path, 'package.json'), 'r'))
author = package['author']
author_name = re.sub('(<(?:.*)>)', '', author).strip()
author_email = re.search(r'<(.*)>', author).groups()[0]

announcement = 'Building for version %s' % package['version']
print('%s' % ''.join(['*'] * len(announcement)))
print(announcement)
print('%s\n\n' % ''.join(['*'] * len(announcement)))

setuptools.setup(
    name=package['name'],
    version=package['version'],
    author=author_name,
    author_email=author_email,
    description=package['description'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/do-py-together/do-py',
    install_requires=[
        'future==0.18.2'
        ],
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    # https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Operating System :: OS Independent',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        ],
    keywords=['development', 'OO']
    )



"""

:date_created: 2019-08-18
"""

import copy

from do_py import DataObject
from do_py.abc import ABCRestrictions


@ABCRestrictions.require('_validate')
class Validator(DataObject):
    """
    Validator provides capability for more complex data validations. Use ManagedRestrictions for simpler single key
    validations and standardizations. In certain DataObjects, the keys are mutually dependent in the sense that the
    value one key stores depends on what the other key(s) are storing. E.g., consider the data object in the example
    below. Cities 'Dallas' and 'Los Angeles' are only valid with 'TX' and 'CA', respectively. 'Dallas' combined with
    'TX' should be rejected. Validator provides a solution for such problems. In some situations, a combination of
    ManagedRestrictions and Validator may be required.

    Users should implement _validate method and include the validation logic.

    NOTE: This is heavy. Use ManagedRestrictions if possible.

    Example:
        class Address(DataObject):
            _restrictions = {
                'city': (['Dallas', 'Los Angeles'], None),
                'state': (['TX', 'CA'], None)
                }

    :attribute _validate: a validation function to validate data for mutually dependent keys
    """
    _is_abstract_ = True

    def __init__(self, data=None, strict=True):
        super(Validator, self).__init__(data=data, strict=strict)
        if strict:
            self._validate()

    def __setitem__(self, key, value):
        """
        A copy of current data is cached before running _validate. In case of exception, the data is restored using
        the copy.
        """
        self_copy = copy.deepcopy(self)
        super(Validator, self).__setitem__(key, value)
        try:
            # Additional validation
            self._validate()
        except:
            self(data=self_copy)
            raise

    def _validate(self):
        """
        Must be implemented by user. Place all validation logic in this method.
        """
        raise NotImplementedError('_validate must be implemented!')



import copy

from do_py.abc import ABCRestrictions, SystemMessages, ABCRestrictionMeta, classproperty
from do_py.data_object.restriction import Restriction
from do_py.exceptions import RestrictionError, DataObjectError
from .restricted_dict import RestrictedDictMixin


@ABCRestrictions.require('_restrictions')
class DataObject(RestrictedDictMixin):
    """
    This is an interface to implement into any data-oriented child class. This is designed to support
    very tight coupling with data and its structure from DB. Data structure and value validation is provided.

    Intended Implementation: use the mix-ins derived from this interface (Loadable, Listable, etc.). Only inherit this
    class directly to extend the data validations on DataObject instantiation, i.e. reporting DataObject
    validates account meta.

    It is important to note the difference between the attribute and key namespaces. All values assigned in the key
    namespace will be validated against the _restrictions defined.


    Example:
        Data Structure:
        ____________________________________________________________________
        | id INT(11) | value TEXT | created TIMESTAMP | modified TIMESTAMP |

        Matching _restrictions value:
        _restrictions = {
            'id': ([int], None),
            'value': ([str, unicode], None),
            'created': ([int, long], None),
            'modified': ([int, long], None)
            }


    :attribute _restrictions: dictionary defining data structure and valid values.
    """
    _schema = None

    @classmethod
    def __compile__(cls):
        """
        This enforces restrictions. We do not want users to instantiate this class.
        """
        assert type(cls._restrictions) is dict, SystemMessages.REQUIRED_FOR % ('_restrictions', cls.__name__)
        for k in cls._restrictions:
            if hasattr(cls, k):
                raise AttributeError('"%s" is already defined in class namespace!' % k)
            try:
                cls._restrictions[k] = Restriction.legacy(cls._restrictions[k])
            except RestrictionError as e:
                raise DataObjectError.from_restriction_error(k, cls, e)

    @classmethod
    def _validate_data(cls, _restrictions, d, strict=True):
        """
        Validate data as per Data Object (DO) restrictions.

        Strict vs. Non-strict:
        In strict initialization, data must contain all the keys required by DO _restrictions.
        In non-strict initialization, it is acceptable to have some keys missing per DO _restrictions. For all missing
        keys, the default restriction value is used.

        Keyspace Integrity:
        Keys that are not specified in _restrictions are not allowed in strict or non-strict initializations.

        Data Integrity:
        Data is validated in both strict and non-strict mode.

        :param _restrictions: DO restrictions
        :type _restrictions: dict
        :param d: Data
        :type d: dict
        :param strict: strict vs non-strict initialization.
        :type strict: bool
        :return: Validated data. Some restrictions may further apply data standardization.
        :rtype: dict
        :raises DataObjectError: When a key not defined in _restrictions is passed in.
        :raises DataObjectError: When an invalid value is passed in.
        """
        _dict = dict()
        d = {} if d is None else d
        # NOTE: Unrestricted keys are never allowed.
        for k in list(d.keys()):
            if k not in _restrictions:
                raise DataObjectError.from_unknown_key(k, cls)

        # NOTE: Use default in strict for missing keys in data.
        for k, v in _restrictions.items():
            if k not in d:
                if strict:
                    raise DataObjectError.from_required_key(k, cls)
                else:
                    _dict[k] = v.default
            else:
                try:
                    _dict[k] = v(d[k], strict=strict)
                except RestrictionError as e:
                    raise DataObjectError.from_restriction_error(k, cls, e)

        return _dict

    def __init__(self, data=None, strict=True):
        """
        Initialize DataObject.
        :param data: Initialize to this dictionary.
        :param strict: See Strict vs Non-strict initialization comments in _validate_data.
        """
        self._strict = strict
        super(DataObject, self).__init__(self._validate_data(self._restrictions, data, strict=strict))
        # NOTE: Now that we are done loading, we go back to strict mode
        self._strict = True

    def __call__(self, data=None, strict=True):
        """
        This re-initializes the data object.
        :param data: Initialize to this dictionary.
        :param strict: See Strict vs Non-strict initialization comments in _validate_data.
        """
        self.__init__(data=data, strict=strict)
        return self

    def __setitem__(self, item, value):
        """
        This assigns a value to item in the key namespace. This value will undergo data validation.
        """
        super(DataObject, self).__setitem__(item, self._restrictions[item](value))

    def __copy__(self):
        """
        Supports shallow copy of DataObject. This gives user back plain old python dictionary.

        NOTE: Allowing this is questionable. Any nested objects will be copied by reference, meaning unintended
        mutation may happen.
        :return: Python dict equivalent of the DataObject
        :rtype: dict
        """
        return dict(self)

    # TODO: Needs a test
    def __deepcopy__(self, memodict={}):
        """
        Supports deep copying of DataObject. This gives user back plain old python dictionary.
        :param memodict:
        :type memodict:
        :return: python dictionary representation
        :rtype: dict
        """
        # NOTE: this could be an alternative implementation
        # return self.__class__(data=dict(self))
        memodict[id(self)] = self.__copy__()
        return copy.deepcopy(self.__copy__())

    @classproperty
    def schema(cls):
        """
        Schema of data object is its key structure based on restrictions. Nested data objects are also supported.
        :return: schema
        :rtype: dict
        """
        if not cls._schema:
            s = dict()
            for k, v in cls._restrictions.items():
                if type(v) is ABCRestrictionMeta:
                    s[k] = v.schema
                elif type(v) is tuple and type(v[0]) is ABCRestrictionMeta:
                    s[k] = v[0].schema
                else:
                    s[k] = None
            cls._schema = s
        return cls._schema

    def __dir__(self):
        return super(DataObject, self).__dir__() + list(self._restrictions.keys())
